import importlib
import logging
from inspect import getfullargspec

from aiohttp import web
from apispec import APISpec
from apispec.ext.marshmallow import OpenAPIConverter

from marshmallow import fields, Schema
from marshmallow.schema import SchemaMeta

from aiohug.arguments import get_default_args
from aiohug.directives import get_available_directives
from .decorators import response, spec

logger = logging.getLogger(__name__)

converter = OpenAPIConverter("2.1")

DEFAULT_HOST = "localhost:8080"
DEFAULT_SCHEMES = ["http"]
DEFAULT_VERSION = None
DEFAULT_TITLE = "Swagger Application"
DEFAULT_DEFINITIONS_PATH = None
DEFAULT_TESTING_MODE = False
DEFAULT_USE_DEFAULT_RESPONSE = True
DEFAULT_DESCRIPTION = None
DEFAULT_CONTACT_EMAIL = None

PARAMETER_IN_PATH = "path"
PARAMETER_IN_QUERY = "query"


def get_summary(doc):
    if doc is not None:
        return doc.split("\n")[0]


def where_is_parameter(name, url):
    return PARAMETER_IN_PATH if "{%s}" % name in url else PARAMETER_IN_QUERY


def get_parameters(url, handler, spec):
    defaults = get_default_args(handler._original_handler)
    args_spec = getfullargspec(handler._original_handler)
    parameters = []
    for name in args_spec.args:
        kind = args_spec.annotations.get(name, fields.Field())
        if name in get_available_directives() or name == "request":
            continue

        if isinstance(kind, fields.Field):
            parameter_place = where_is_parameter(name, url)
            kind.metadata = {"location": where_is_parameter(name, url)}
            kind.required = name not in defaults
            parameter = converter.field2parameter(
                kind, name=name, default_in=parameter_place, use_refs=False
            )
            if name in defaults:
                parameter["default"] = defaults[name]
            parameters.append(parameter)
        # body
        elif name == "body" and (
            isinstance(kind, Schema) or isinstance(kind, SchemaMeta)
        ):
            if isinstance(kind, Schema):
                schema_name = kind.__class__.__name__
                schema = kind
            elif isinstance(kind, SchemaMeta):
                schema_name = kind.__name__
                schema = kind()

            spec.definition(schema_name, schema=schema)

            ref_definition = "#/definitions/{}".format(schema_name)
            ref_schema = {"$ref": ref_definition}

            parameters.append(
                {"in": "body", "name": "body", "required": True, "schema": ref_schema}
            )

    return parameters


def generate_swagger(
    app: web.Application,
    title=DEFAULT_TITLE,
    version=DEFAULT_VERSION,
    host=DEFAULT_HOST,
    schemes=DEFAULT_SCHEMES,
    definitions_path=DEFAULT_DEFINITIONS_PATH,
    **options
):
    if host is not None:
        options["host"] = host
    if schemes is not None:
        options["schemes"] = schemes

    spec = APISpec(
        title=title, version=version, plugins=("apispec.ext.marshmallow",), **options
    )
    if definitions_path is not None:
        definitions = importlib.import_module(definitions_path)

        for name, schema in definitions.__dict__.items():  # type: str, Schema
            if name.endswith("Schema") and len(name) > len("Schema"):
                spec.definition(name, schema=schema)

    resources = app.router._resources
    for resource in resources:
        url = resource.canonical
        name = resource.name
        for route in resource._routes:
            method = route.method
            if method == "HEAD":
                continue

            handler = route._handler

            try:
                handler_spec = handler.swagger_spec
            except AttributeError:
                handler_spec = {}

            if "excluded" in handler_spec:
                continue

            try:
                handler_spec["summary"] = get_summary(handler.__doc__)
                handler_spec["description"] = handler.__doc__
            except KeyError:
                pass

            parameters = get_parameters(url, handler, spec)
            if parameters:
                handler_spec["parameters"] = parameters

            handler_spec["operationId"] = name

            spec.add_path(url, operations={method.lower(): handler_spec})

    return spec.to_dict()



import pytest
from marshmallow import Schema, fields

from aiohug import swagger
from aiohug.swagger.decorators import ensure_swagger_attr


class TestSchema(Schema):
    field = fields.Integer()


def test_ensure_swagger_attr():
    def handler():
        pass

    with pytest.raises(AttributeError):
        handler.swagger_spec

    ensure_swagger_attr(handler)
    handler.swagger_spec == {"responses": {}}


def test_response():
    code = 201
    schema = TestSchema
    description = "test"

    @swagger.response(code, schema=schema, description=description)
    def handler():
        pass

    assert code in handler.swagger_spec["responses"]
    assert handler.swagger_spec["responses"][code]["schema"] == schema
    assert handler.swagger_spec["responses"][code]["description"] == description


def test_response_code():
    code = 201

    @swagger.response(code)
    def handler():
        pass

    assert code in handler.swagger_spec["responses"]
    assert handler.swagger_spec["responses"][code] == {}


def test_spec():
    attrs = {"private": True, "exclude": True, "deprecated": True, "tags": ["test"]}

    @swagger.spec(**attrs)
    def handler():
        pass

    for attr, value in attrs.items():
        assert handler.swagger_spec[attr] == value


def test_spec_response_codes():
    codes = [200, 409]

    @swagger.spec(response_codes=codes)
    def handler():
        pass

    assert list(handler.swagger_spec["responses"].keys()) == codes



import pytest

from aiohug.swagger import get_summary, where_is_parameter


def test_get_summary():
    doc = """Some string
    
    Hello
    """
    assert get_summary(doc) == "Some string"


def test_get_summary_for_empty_string():
    assert get_summary("") == ""


def test_get_summary_for_none():
    assert get_summary(None) == None


@pytest.mark.parametrize(
    "name,url,place",
    (
        ("page_id", "https://example.com/page/{page_id}?published=False", "path"),
        ("page_id", "https://example.com/page/?page_id=5&published=False", "query"),
    ),
)
def test_where_is_parameter(name, url, place):
    assert where_is_parameter(name, url) == place



from typing import Optional, Iterable


def ensure_swagger_attr(handler):
    try:
        handler.swagger_spec
    except AttributeError:
        handler.swagger_spec = {"responses": {}}


def response(response_code, schema=None, description=None):
    """A decorator that adds swagger response"""

    def decorator(handler):
        ensure_swagger_attr(handler)
        handler.swagger_spec["responses"][response_code] = {}
        if schema is not None:
            handler.swagger_spec["responses"][response_code]["schema"] = schema
        if description is not None:
            handler.swagger_spec["responses"][response_code][
                "description"
            ] = description
        return handler

    return decorator


def spec(
    exclude: bool = False,
    private: bool = False,
    deprecated: bool = False,
    tags: Optional[Iterable] = None,
    response_codes: Optional[Iterable] = None,
):
    def decorator(handler):
        ensure_swagger_attr(handler)
        handler.swagger_spec["private"] = private
        handler.swagger_spec["exclude"] = exclude
        handler.swagger_spec["deprecated"] = deprecated
        if tags is not None:
            handler.swagger_spec["tags"] = tags
        if response_codes is not None:
            for code in response_codes:
                handler.swagger_spec["responses"][code] = {}

        return handler

    return decorator
